from scalecast.Forecaster import Forecaster
from scalecast.MVForecaster import MVForecaster
from scalecast.util import break_mv_forecaster, find_optimal_lag_order, find_optimal_coint_rank
from scalecast.auxmodels import mlp_stack
import pandas_datareader as pdr
import matplotlib.pyplot as plt
import pickle
from sklearn.preprocessing import RobustScaler

def build_MVForecaster(test_length=24):
    s1 = pdr.get_data_fred('UTUR',start='2000-01-01',end='2022-10-01')
    s2 = pdr.get_data_fred('UNRATE',start='2000-01-01',end='2022-10-01')
    s3 = pdr.get_data_fred('SAHMREALTIME',start='2000-01-01',end='2022-10-01')

    f1 = Forecaster(y=s1['UTUR'],current_dates=s1.index,future_dates=24)
    f2 = Forecaster(y=s2['UNRATE'],current_dates=s2.index,future_dates=24)
    f3 = Forecaster(y=s3['SAHMREALTIME'],current_dates=s3.index,future_dates=24)

    return MVForecaster(
        f1,
        f2,
        f3,
        test_length = test_length,
        names=['UTUR','UNRATE','SAHMREALTIME'],
        merge_Xvars='i',
    )

def weighted_series(x):
    return x[0]*.75+x[1]*.25

def test_pickle():
    mvf = build_MVForecaster()
    mvf.set_optimize_on(weighted_series)
    mvf.add_optimizer_func(weighted_series,'weighted')
    mvf.set_optimize_on('weighted')
    with open('../../mvf.pckl','wb') as pckl:
        pickle.dump(mvf,pckl)

def test_modeling():
    for tl in (0,36):
        mvf = build_MVForecaster(test_length = tl)
        mvf.add_normalizer('robust',RobustScaler)
        mvf.set_metrics(['rmse','r2'])
        mvf
        if tl > 0:
            mvf.eval_cis(cilevel = .9)

        models = ('lasso','xgboost')

        mvf.corr_lags('UNRATE','UTUR',lags=5)
        mvf.corr_lags('UNRATE','UTUR',lags=5,disp='heatmap',annot=True,vmin=-1,vmax=1)
        plt.savefig('../../corr_lags.png')
        plt.close()
        find_optimal_lag_order(mvf)
        find_optimal_coint_rank(mvf,det_order=-1,k_ar_diff=8)
        mvf.tune_test_forecast(
            models,
            limit_grid_size=.2,
            min_grid_size=4,
            cross_validate = True,
            test_length = 36,
            train_length = 500,
            space_between_sets = 12,
            rolling = True,
            k = 2,
            dynamic_tuning = 24,
            error = 'raise',
            suffix = '_cv',
        )
        mvf.set_estimator('vecm')
        mvf.set_grids_file('VECMGrid')
        mvf.cross_validate(k=2)
        mvf.auto_forecast()

        if tl > 0:
            mlp_stack(
                mvf,
                model_nicknames=['xgboost_cv','lasso_cv'],
                lags = 6,
            )

        mvf.set_best_model(
            determine_best_by = (
                'ValidationMetricValue' if tl == 0 
                else 'TestSetRMSE'
            )
        )

        mvf.plot(ci=True)
        plt.savefig(f'../../mvf_plot_{tl}.png')
        plt.close()

        if tl > 0:
            mvf.plot_test_set(ci=True)
            plt.savefig('../../mvf_plot_ts.png')
            plt.close()

        mvf.plot_fitted()
        plt.savefig(f'../../mvf_fvs_{tl}.png')
        plt.close()

        mvf.export(to_excel=True,out_path='../..',excel_name=f'mv_results_{tl}.xlsx',cis=True)
        mvf.export_fitted_vals().to_excel('../../mv_fvs.xlsx',index=False)
        with open('../../mvf.pckl','wb') as pckl:
            pickle.dump(mvf,pckl)

        f1, f2, f3 = break_mv_forecaster(mvf)

def main():
    test_pickle()
    test_modeling()

if __name__ == '__main__':
    main()


from __future__ import annotations
from .cfg import COLORS, SERIES_COLORS, IGNORE_AS_HYPERPARAMS, MV_ESTIMATORS
from ._utils import _developer_utils, _tune_test_forecast
from ._Forecaster_parent import Forecaster_parent, ForecastError
from .types import (
    PositiveInt, 
    NonNegativeInt, 
    ConfInterval, 
    ModelValues,
    DetermineBestBy,
    EvaluatedModel,
    AvailableModel,
    ExportOptions,
    SeriesName,
    SeriesValues,
)
from .classes import AR, EvaluatedMetric
from typing import Optional, Literal, Any, Self, TYPE_CHECKING
import warnings
import os
from pathlib import Path
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import seaborn as sns
from itertools import cycle
import warnings
import datetime
if TYPE_CHECKING:
    from .Forecaster import Forecaster

class MVForecaster(Forecaster_parent):
    """ MVForecaster is a class for forecasting multiple series at once with the same models and hyperparameters.

    Args:
        *fs (Forecaster): Forecaster objects
        names (list-like): Optional. An array with the same number of elements as *fs that can be used to map to each series.
            Ex. if names == ['UTUR','UNRATE'], the user must now refer to the series with the selected names. 
            If specific names are not supplied, refer to the series with y1, y2, etc.
            The order the series are supplied will be maintained.
        not_same_len_action (str): One of 'trim', 'fail'. default 'trim'.
            What to do with series that are different lengths.
            'trim' will trim each series so that all dates line up.
        merge_Xvars (str): One of 'union', 'u', 'intersection', 'i'. default 'union'.
            How to combine Xvars in each object.
            'union' or 'u' combines all regressors from each object.
            'intersection' or 'i' combines only regressors that all objects have in common.
        merge_future_dates (str): One of 'longest', 'shortest'. Default 'longest'.
            Which future dates to use in the various series. This can be changed later.
        test_length (int or float): Default 0. The test length that all models will use to test all models out of sample.
            If float, must be between 0 and 1 and will be treated as a fractional split.
            By default, models will not be tested.
        validation_length (int): The size of the validation set. Default sets to 1.
        metrics (list[str]): Optional. List of metrics to evaluate every model. 
        optimize_on (str): The way to aggregate the derived metrics when optimizing models across all series. 
            This can be a function: 'mean', 'min', 'max', a custom function that takes a list of objects and returns an aggregate function (such as a weighted average) 
            or a series name. Custom functions and weighted averages can also be added later
            by calling mvf.set_optimize_on().
        cis (bool): Default False. Whether to evaluate probabilistic confidence intervals for every model evaluated.
            If setting to True, ensure you also set a test_length of at least 20 observations for 95% confidence intervals.
            See eval_cis() and set_cilevel() methods and docstrings for more information.
        carry_fit_models (bool): Default False.
            Whether to store the regression model for each fitted model in history.
            Setting this to False can save memory.
    """
    def __init__(
        self,
        *fs:'Forecaster',
        names:Optional[str]=None,
        not_same_len_action:Literal["trim","fail"]="trim",
        merge_Xvars:Literal["u","union","intersection","i"]="union",
        merge_future_dates:Literal["longest","shortest"]="longest",
        test_length:NonNegativeInt = 0,
        validation_length:NonNegativeInt=1,
        metrics:Optional[list[str]]=None,
        optimize_on:Literal['mean','min','max']|callable|SeriesName = 'mean',
        cis:bool = False,
        carry_fit_models:bool = False,
    ):
        super().__init__(
            y = fs[0].y, # placeholder -- will be overwritten
            test_length = test_length,
            cis = cis,
            validation_length=validation_length,
        )
        for f in fs:
            f._typ_set()
        if (
            len(set([len(f.current_dates) for f in fs])) > 1
            or len(set([min(f.current_dates) for f in fs])) > 1
        ):
            if not_same_len_action == "fail":
                raise ValueError("All series must be same length.")
            elif not_same_len_action == "trim":
                from .multiseries import line_up_dates
                line_up_dates(*fs)
            else:
                raise ValueError(f'not_same_len_action must be one of ("trim","fail"), got {not_same_len_action}.')

        if len(set([f.freq for f in fs])) > 1:
            raise ValueError("All date frequencies in passed Forecaster objects must be equal.")
        if len(fs) < 2:
            raise ValueError("Must pass at least two series.")

        self.grids_file = 'MVGrids'
        self.freq = fs[0].freq
        self.n_series = len(fs)
        self.estimators = MV_ESTIMATORS

        self.y = {}
        if names is None:
            names = [f'y{i+1}' for i in range(self.n_series)]
        self.names = names
        for i, f in enumerate(fs):
            if i == 0:
                self.current_dates = f.current_dates.copy().reset_index(drop=True)
            self.y[names[i]] = f.y.copy()
            if merge_Xvars in ("union", "u"):
                if i == 0:
                    self.current_xreg = {
                        k: v.copy().reset_index(drop=True)
                        for k, v in f.current_xreg.items()
                        if not isinstance(k,AR)
                    }
                    self.future_xreg = {
                        k: v[:]
                        for k, v in f.future_xreg.items()
                        if not isinstance(k,AR)
                    }
                else:
                    for k, v in f.current_xreg.items():
                        if not isinstance(k,AR):
                            self.current_xreg[k] = v.copy().reset_index(drop=True)
                            self.future_xreg[k] = f.future_xreg[k][:]
            elif merge_Xvars in ("intersection", "i"):
                if i == 0:
                    self.current_xreg = {
                        k: v.copy().reset_index(drop=True) for k,v in f.current_xreg.items() if not isinstance(k,AR)
                    }
                    self.future_xreg = {k: v[:] for k,v in f.future_xreg.items() if not isinstance(k,AR)}
                else:
                    f.drop_Xvars(*[k for k in f.current_xreg if k not in self.current_xreg or isinstance(k,AR)])

            else:
                raise ValueError(
                    f"merge_Xvars must be one of ('union','u','intersection','i'), got {merge_Xvars}"
                )

        self.optimizer_funcs = {
            "mean": np.mean,
            "min": np.min,
            "max": np.max,
        }
        self.set_optimize_on(optimize_on)
        future_dates_lengths = {i: len(f.future_dates) for i, f in enumerate(fs)}
        if merge_future_dates == "longest":
            self.future_dates = [
                fs[i].future_dates
                for i, v in future_dates_lengths.items()
                if v == max(future_dates_lengths.values())
            ][0]
        elif merge_future_dates == "shortest":
            self.future_dates = [
                fs[i].future_dates
                for i, v in future_dates_lengths.items()
                if v == min(future_dates_lengths.values())
            ][0]
        else:
            raise ValueError(f"merge_future_dates must be one of ('longest','shortest'), got {merge_future_dates}.")
        if metrics:
            self.set_metrics(metrics)
        self.carry_fit_models = carry_fit_models

    def __repr__(self):
        return """MVForecaster(
    DateStartActuals={}
    DateEndActuals={}
    Freq={}
    N_actuals={}
    N_series={}
    SeriesNames={}
    ForecastLength={}
    Xvars={}
    TestLength={}
    ValidationLength={}
    ValidationMetric={}
    ForecastsEvaluated={}
    CILevel={}
    CurrentEstimator={}
    OptimizeOn={}
    GridsFile={}
)""".format(
        self.current_dates.values[0].astype(str),
        self.current_dates.values[-1].astype(str),
        self.freq,
        self.n_actuals(),
        self.n_series,
        self.names,
        len(self.future_dates),
        list(self.current_xreg.keys()),
        self.test_length,
        self.validation_length,
        self.validation_metric,
        list(self.history.keys()),
        self.cilevel if self.cis is True else None,
        self.estimator,
        self.optimize_on,
        self.grids_file,
    )

    def add_optimizer_func(self, func:callable, called:Optional[str] = None) -> Self:
        """ Add an optimizer function that can be used to determine the best-performing model.
        This is in addition to the 'mean', 'min', and 'max' functions that are available by default.

        Args:
            func (Function): The function to add.
            called (str): Optional. How to refer to the function when calling `optimize_on()`.
                If left unspecified, will use the name of the function.

        Returns:
            Self

        >>> def weighted(x):
        >>>     # weighted average of first two series in the object
        >>>     return x[0]*.25 + x[1]*.75
        >>> mvf.add_optimizer_func(weighted)
        >>> mvf.set_optimize_on('weighted') # optimize on that function
        >>> mvf.set_estimator('mlr')
        >>> mvf.tune() # best model now chosen based on the weighted average function you added; series2 gets 3x the weight of series 1
        """
        called = self._called(func,called)
        self.optimizer_funcs[called] = func
        return self

    def _typ_set(self):
        """ Placeholder function.
        """
        return

    def add_signals(
        self,
        model_nicknames:list[EvaluatedModel], 
        series:SeriesName|Literal['all'] = 'all', 
        fill_strategy:Literal['actuals','bfill']|None = 'actuals', 
        train_only:bool = False
    ) -> Self:
        """ Adds the predictions from already-evaluated models as covariates that can be used for future evaluated models.
        The names of the added variables will all begin with "signal_" and end with the given model nickname folowed by the series name.

        Args:
            model_nicknames (list): The names of already-evaluated models with information stored in the history attribute.
            fill_strategy (str or None): The strategy to fill NA values that are present at the beginning of a given model's fitted values.
                Available options are: 'actuals' (default) which will replace nulls with actuals;
                'bfill' which will backfill null values;
                or None which will leave null values alone, which can cause errors in future evaluated models.
            train_only (bool): Default False. Whether to add fitted values from the training set only.
                The test-set predictions will be out-of-sample if this is True. The future unknown values are always out-of-sample.
                Even when this is True, the future unknown values are taken from a model trained on the full set of
                known observations.

        Returns:
            Self

        >>> mvf.set_estimator('xgboost')
        >>> mvf.manual_forecast()
        >>> mvf.add_signals(model_nicknames = ['xgboost']) # adds regressors called 'signal_xgboost_{series1name}', ..., 'signal_xgboost_{seriesNname}'
        """
        series = self._parse_series(series)
        for m in model_nicknames:
            for s in series:
                fcst = self.history[m]['Forecast'][s][:]
                fvs = self.history[m]['FittedVals'][s][:]
                num_fvs = len(fvs)
                match fill_strategy:
                    case 'actuals':
                        pad = [np.nan if fill_strategy is None else fvs[0]] * (len(self.y[s]) - num_fvs)
                    case _:
                        pad = self.y[s].to_list()[:-num_fvs]

                self.current_xreg[f'signal_{m}_{s}'] = pd.Series(pad + fvs)
                
                if train_only:
                    tsp = self.history[m]['TestSetPredictions'][s][:]
                    self.current_xreg[f'signal_{m}_{s}'].iloc[-len(tsp):] = tsp
                
                self.future_xreg[f'signal_{m}_{s}'] = fcst
        
        return self

    def chop_from_front(self,n:NonNegativeInt,fcst_length:Optional[NonNegativeInt]=None) -> Self:
        """ Cuts the amount of y observations in the object from the front counting backwards.
        The current length of the forecast horizon will be maintained and all future regressors will be rewritten to the appropriate attributes.

        Args:
            n (int):
                The number of observations to cut from the front.
            fcst_length (int): Optional.
                The new length of the forecast length.
                By default, maintains the same forecast length currently in the object.

        Returns:
            Self

        >>> mvf.chop_from_front(10) # keeps all observations before the last 10
        """
        n = int(n)
        fcst_length = len(self.future_dates) if fcst_length is None else fcst_length
        self.y = {k:v.iloc[:-n] for k, v in self.y.items()}
        self.current_dates = self.current_dates.iloc[:-n]
        self.generate_future_dates(fcst_length)
        self.future_xreg = {
            k:(self.current_xreg[k].to_list()[-n:] + v[:max(0,(fcst_length-n))])[-fcst_length:]
            for k, v in self.future_xreg.items()
        }
        self.current_xreg = {k:v.iloc[:-n].reset_index(drop=True) for k, v in self.current_xreg.items()}
        return self

    def keep_smaller_history(self, n:NonNegativeInt) -> Self:
        """ Cuts y observations in the object by counting back from the beginning.

        Args:
            n (int, str, or datetime.datetime):
                If int, the number of observations to keep.
                Otherwise, the last observation to keep.
                Must be parsable by pandas' Timestamp function.

        Returns:
            Self

        >>> f.keep_smaller_history(500) # keeps last 500 observations
        >>> f.keep_smaller_history('2020-01-01') # keeps only observations on or later than 1/1/2020
        """
        if (type(n) is datetime.datetime) or (type(n) is pd.Timestamp) or isinstance(n,str):
            n = len([i for i in self.current_dates if i >= pd.Timestamp(n)])
        
        n = int(n)
        _developer_utils.descriptive_assert(
            isinstance(n, int),
            ValueError,
            "n must be an int, datetime object, or str and there must be more than 2 observations to keep.",
        )
        _developer_utils.descriptive_assert(
            n > 2,
            ValueError,
            "n must be an int, datetime object, or str and there must be more than 2 observations to keep.",
        )
        self.y = {k: v.iloc[-n:] for k, v in self.y.items()}
        self.current_dates = self.current_dates.iloc[-n:]
        self.current_xreg = {k:v.iloc[-n:].reset_index(drop=True) for k, v in self.current_xreg.items()}
        return self

    def tune_test_forecast(
        self,
        models:list[AvailableModel],
        cross_validate:bool=False,
        dynamic_tuning:bool=False,
        dynamic_testing:bool=True,
        limit_grid_size:Optional[NonNegativeInt|ConfInterval]=None,
        min_grid_size:NonNegativeInt=1,
        suffix:Optional[str]=None,
        error:Literal['ignore','raise','warn']='raise',
        **cvkwargs:Any,
    ) -> Self:
        """ Iterates through a list of models, tunes them using grids in a grids file, and forecasts them.

        Args:
            models (list-like): The models to iterate through.
            cross_validate (bool): Default False.
                Whether to tune the model with cross validation. 
                If False, uses the validation slice of data to tune.
            dynamic_tuning (bool or int): Default False.
                Whether to dynamically tune the model or, if int, how many forecast steps to dynamically tune it.
            dynamic_testing (bool or int): Default True.
                Whether to dynamically/recursively test the forecast (meaning AR terms will be propagated with predicted values).
                If True, evaluates dynamically over the entire out-of-sample slice of data.
                If int, window evaluates over that many steps (2 for 2-step dynamic forecasting, 12 for 12-step, etc.).
                Setting this to False or 1 means faster performance, 
                but gives a less-good indication of how well the forecast will perform more than one period out.
                The model will skip testing if the test_length attribute is set to 0.
            limit_grid_size (int or float): Optional. Pass an argument here to limit each of the grids being read.
                See https://scalecast.readthedocs.io/en/latest/Forecaster/MVForecaster.html#src.scalecast.MVForecaster.MVForecaster.limit_grid_size.
            min_grid_size (int): Default 1. The smallest grid size to keep. Ignored if limit_grid_size is None.
            suffix (str): Optional. A suffix to add to each model as it is evaluated to differentiate them when called later. 
                If unspecified, each model can be called by its estimator name.
            error (str): One of 'ignore','raise','warn'. Default 'raise'.
                What to do with the error if a given model fails.
                'warn' logs a warning that the model could not be evaluated.
            **cvkwargs: Passed to the cross_validate() method.

        Returns:
            Self

        >>> models = ('mlr','mlp','lightgbm')
        >>> mvf.tune_test_forecast(models,dynamic_testing=False)
        """
        _tune_test_forecast(
            f = self,
            models = models,
            cross_validate=cross_validate,
            dynamic_tuning=dynamic_tuning,
            dynamic_testing=dynamic_testing,
            limit_grid_size=limit_grid_size,
            min_grid_size=min_grid_size,
            suffix=suffix,
            error=error,
            **cvkwargs,
        )
        return self

    def set_optimize_on(self, how:str) -> Self:
        """ Choose how to determine best models by choosing which series should be optimized or the aggregate function to apply on the derived metrics across all series.
        This is the decision that will be used for optimizing model hyperparameters.

        Args:
            how (str): One of MVForecaster.optimizer_funcs, a series name, or a function. 
                Only one series name will be in mvf.optimizer_funcs at a given time.
                mvf.optimize_on is set to 'mean' when the object is initiated.
        
        Returns:
            Self
        """
        if callable(how):
            self.add_optimizer_func(how)
            self.optimize_on = how.__name__
            return
        elif how in self.names: # assume it's a series name
            if self.optimize_on == how:
                return
            elif np.any([o in self.names for o in self.optimizer_funcs]):
                to_pop = [o for o in self.optimizer_funcs if o in self.names]
                for p in to_pop:
                    self.optimizer_funcs.pop(p)
            self.series_to_optimize = self.names.index(how)
            self.add_optimizer_func(func=lambda x: self.series_to_optimize,called=how)
        elif how not in self.optimizer_funcs:
            raise ValueError(
                f'Value passed to how cannot be used: {how}. '
                f'Possible values are: {list(self.optimizer_funcs.keys())} or a function.'
            )

        self.optimize_on = how
        return self

    def _set_cis(self,*attrs,m,ci_range,preds):
        for i, attr in enumerate(attrs):
            self.history[m][attr] = {
                k:p + (ci_range[k] if i%2 == 0 else (ci_range[k]*-1))
                for k,p in preds.items()
            }

    def _bank_history(self, **kwargs:Any):
        """ places all relevant information from the last evaluated forecast into the history dictionary attribute
            **kwargs: passed from each model, depending on how that model uses Xvars, normalizer, and other args
        """
        call_me = self.call_me
        fitted_val_actuals = {k: (v.to_list()[-len(self.fitted_values[k]):]) for k, v in self.y.items()}
        
        self.history[call_me]['Estimator'] = self.estimator
        if hasattr(self.call_estimator,'Xvars'):
            self.history[call_me]['Xvars'] = self.call_estimator.Xvars
        else:
            self.history[call_me]['Xvars'] = None
        self.history[call_me]['HyperParams'] = {k: v for k, v in kwargs.items() if k not in IGNORE_AS_HYPERPARAMS}
        self.history[call_me]['Lags'] =  self.call_estimator.lags
        self.history[call_me]['Forecast'] = self.forecast
        self.history[call_me]['Observations'] = len(self.current_dates)
        self.history[call_me]['FittedVals'] = self.fitted_values
        self.history[call_me]['DynamicallyTested'] = self.dynamic_testing
        self.history[call_me]['CILevel'] = self.cilevel if self.cis else np.nan
        if self.carry_fit_models:
            self.history[call_me]['regr'] = self.trained_models
        for attr in ('TestSetPredictions','TestSetActuals'):
            if attr not in self.history[call_me]:
                self.history[call_me][attr] = {n:[] for n in self.names}

        if hasattr(self,'best_params'):
            if np.all([k in self.best_params for k in self.history[call_me]['HyperParams']]):
                self.history[call_me]['ValidationMetric'] = self.validation_metric
                self.history[call_me]['ValidationMetricValue'] = self.validation_metric_value
                self.history[call_me]['grid'] = self.grid
                self.history[call_me]['grid_evaluated'] = self.grid_evaluated

        for met in self.metrics:
            self.history[call_me]["InSample" + met.name.upper()] = {
                series: EvaluatedMetric(store=met,score=_developer_utils._return_na_if_len_zero(a, self.fitted_values[series], met.eval_func))
                for series, a in fitted_val_actuals.items()
            }

        if self.cis and self.history[call_me]['TestSetPredictions'][self.names[0]]:
            self._check_right_test_length_for_cis(self.cilevel)
            fcst = self.history[call_me]['Forecast']
            test_preds = self.history[call_me]['TestSetPredictions']
            test_actuals = self.history[call_me]['TestSetActuals']
            test_resids = {k:np.array(p) - np.array(test_actuals[k]) for k, p in test_preds.items()}
            ci_range = {k:np.percentile(np.abs(r), 100 * self.cilevel) for k,r in test_resids.items()}
            self._set_cis(
                "UpperCI",
                "LowerCI",
                m = call_me,
                ci_range = ci_range,
                preds = fcst,
            )
            self._set_cis(
                "TestSetUpperCI",
                "TestSetLowerCI",
                m = call_me,
                ci_range = ci_range,
                preds = test_preds,
            )

    def set_best_model(self, model:Optional[EvaluatedModel]=None, determine_best_by:Optional[DetermineBestBy]=None) -> Self:
        """ Sets the best model to be referenced as "best".
        One of model or determine_best_by parameters must be specified.

        Args:
            model (str): The model to set as the best.
                Must match the estimator name or call_me if that was used when evaluating the model.
            determine_best_by (str): One of MVForecaster.determine_best_by.
                If model is specified, this will be ignored.

        Returns:
            Self
        """
        if model is not None:
            if model in self.history.keys():
                self.best_model = model
            else:
                raise ValueError(f"Cannot find {model} in history.")
        else:
            self.best_model = self.order_fcsts(determine_best_by=determine_best_by)[0]
        return self

    def _parse_series(self, series:SeriesValues) -> list[SeriesName]:
        match series:
            case 'all':
                return list(self.y.keys())
            case str():
                return [series]
            case _:
                return list(series) 

    def _parse_models(self, models:ModelValues, put_best_on_top:bool) -> list[EvaluatedModel]:
        match models:
            case 'all':
                models = list(self.history.keys())
            case str():
                models = [models]
            case _:
                models = list(models)

        if put_best_on_top:
            models = ([self.best_model] if self.best_model in models else []) + [
                m for m in models if m != self.best_model
            ]
        return models

    def plot(
        self, 
        models:ModelValues="all", 
        series:SeriesValues="all", 
        put_best_on_top:bool=False, 
        ci:bool=False, 
        ax:Axes=None,
        figsize:tuple[int,int]=(12,6),
        colors:Optional[list[str]] = COLORS,
        series_colors:Optional[list[str]] = SERIES_COLORS,
    ) -> Axes:
        """ Plots all forecasts with the actuals, or just actuals if no forecasts have been evaluated or are selected.

        Args:
            models (list-like or str): Default 'all'.
               The forecasted models to plot.
               Name of the model, 'all', or list-like of model names.
               'top_' and None not supported.
            series (list-like or str): Default 'all'.
               The series to plot.
               Name of the series, 'all', or list-like of series names.
            put_best_on_top (bool): Only set to True if you have previously called set_best_model().
                If False, ignored.
            ci (bool): Default False.
                Whether to display the confidence intervals.
            ax (Axis): Optional. The existing axis to write the resulting figure to.
            figsize (tuple): Default (12,6). Size of the resulting figure. Ignored when ax is not None.
            colors (list[str]): Optional. The colors to use when drawing the forecasts.
            series_colors (list[str]): Optional. The colors to use when drawing the actual series.

        Returns:
            (Axis): The figure's axis.

        >>> mvf.plot() # plots all forecasts and all series
        >>> plt.show()
        """
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        colors = cycle(colors)
        series_colors = cycle(series_colors)

        series = self._parse_series(series)
        models = self._parse_models(models, put_best_on_top)

        k = 0
        for s in series:
            series_color = next(series_colors)
            sns.lineplot(
                x=self.current_dates.to_list(),
                y=self.y[s].to_list()[-len(self.current_dates) :],
                label=f"{s} actuals",
                ax=ax,
                color=series_color,
            )
            for m in models:
                color = next(colors)
                sns.lineplot(
                    x=self.future_dates.to_list(),
                    y=self.history[m]["Forecast"][s],
                    label=f"{s} {m}",
                    color=color,
                    ax=ax,
                )

                if ci:
                    try:
                        ax.fill_between(
                            x=self.future_dates.to_list(),
                            y1=self.history[m]["UpperCI"][s],
                            y2=self.history[m]["LowerCI"][s],
                            alpha=0.2,
                            color=color,
                            label="{} {} {:.0%} CI".format(
                                s, m, self.history[m]["CILevel"]
                            ),
                        )
                    except KeyError:
                        _developer_utils._warn_about_not_finding_cis(m)
                k += 1

        ax.legend()
        ax.set_xlabel("Date")
        ax.set_ylabel("Values")
        return ax

    def plot_test_set(
        self,
        models:ModelValues="all",
        series:SeriesValues="all",
        put_best_on_top:bool=False,
        include_train:bool|PositiveInt=True,
        ci:bool=False,
        ax:Axes=None,
        figsize:tuple[int,int]=(12,6),
        colors:Optional[list[str]] = COLORS,
        series_colors:Optional[list[str]] = SERIES_COLORS,
    ) -> Axes:
        """  Plots all test-set predictions with the actuals.

        Args:
            models (list-like or str): Default 'all'.
               The forecasted models to plot.
               Name of the model, 'all', or list-like of model names.
               'top_' and None not supported.
            series (list-like or str): Default 'all'.
               The series to plot.
               Name of the series ('y1...', 'series1...' or user-selected name), 'all', or list-like of series names.
            put_best_on_top (bool): Only set to True if you have previously called set_best_model().
                If False, ignored.
            include_train (bool or int): Default True.
                Use to zoom into training results.
                If True, plots the test results with the entire history in y.
                If False, matches y history to test results and only plots this.
                If int, plots that length of y to match to test results.
            ci (bool): Default False.
                Whether to display the confidence intervals.
            ax (Axis): Optional. The existing axis to write the resulting figure to.
            figsize (tuple): Default (12,6). Size of the resulting figure. Ignored when ax is not None.
            colors (list[str]): Optional. The colors to use when drawing the forecasts.
            series_colors (list[str]): Optional. The colors to use when drawing the actual series.

        Returns:
            (Axis): The figure's axis.

        >>> mvf.plot_test_set() # plots all test set predictions on all series
        >>> plt.show()
        """
        _developer_utils.descriptive_assert(
            self.test_length > 0,
            ForecastError,
            'plot_test_set() does not work when models were not tested (test_length set to 0).',
        )

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        colors = cycle(colors)
        series_colors = cycle(series_colors)

        series = self._parse_series(series)
        models = self._parse_models(models, put_best_on_top)

        match include_train:
            case True:
                include_train = self.n_actuals()
            case False:
                include_train = self.test_length

        k = 0
        for s in series:
            series_color = next(series_colors)
            y = self.y[s].to_list()[-len(self.current_dates) :]
            sns.lineplot(
                x=self.current_dates.to_list()[-include_train:],
                y=y[-include_train:],
                label=f"{s} actual",
                ax=ax,
                color=series_color,
            )
            for m in models:
                color = next(colors)
                sns.lineplot(
                    x=self.current_dates.to_list()[-self.test_length :],
                    y=self.history[m]["TestSetPredictions"][s],
                    label=f"{s} {m}",
                    color=color,
                    linestyle="--",
                    alpha=0.7,
                    ax=ax,
                )

                if ci:
                    try:
                        plt.fill_between(
                            x=self.current_dates.to_list()[-self.test_length :],
                            y1=self.history[m]["TestSetUpperCI"][s],
                            y2=self.history[m]["TestSetLowerCI"][s],
                            alpha=0.2,
                            color=color,
                            label="{} {} {:.0%} CI".format(
                                s, m, self.history[m]["CILevel"]
                            ),
                        )
                    except KeyError:
                        _developer_utils._warn_about_not_finding_cis(m)
                k += 1

        ax.legend()
        ax.set_xlabel("Date")
        ax.set_ylabel("Values")
        return ax

    def plot_fitted(
        self, 
        models:ModelValues="all", 
        series:SeriesValues="all", 
        ax:Axes=None,
        figsize:tuple[int,int]=(12,6),
        colors:Optional[list[str]] = COLORS,
        series_colors:Optional[list[str]] = SERIES_COLORS,
    ) -> Axes:
        """ Plots fitted values with the actuals.

        Args:
            models (list-like or str): Default 'all'.
               The forecasted models to plot.
               Name of the model, 'all', or list-like of model names.
               'top_' and None not supported.
            series (list-like or str): Default 'all'.
               The series to plot.
               Name of the series ('y1...', 'series1...' or user-selected name), 'all', or list-like of series names.
            ax (Axis): Optional. The existing axis to write the resulting figure to.
            figsize (tuple): Default (12,6). Size of the resulting figure. Ignored when ax is not None.
            colors (list[str]): Optional. The colors to use when drawing the forecasts.
            series_colors (list[str]): Optional. The colors to use when drawing the actual series.

        Returns:
            (Axis): The figure's axis.

        >>> mvf.plot_fitted() # plots all fitted values on all series
        >>> plt.show()
        """
        series = self._parse_series(series)
        models = self._parse_models(models, False)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        colors = cycle(colors)
        series_colors = cycle(series_colors)

        k = 0
        for s in series:
            series_color = next(series_colors)
            act = self.y[s].to_list()
            sns.lineplot(
                x=self.current_dates.to_list(),
                y=act[-len(self.current_dates):],
                label=f"{s} actual",
                ax=ax,
                color=series_color,
            )
            for m in models:
                color = next(colors)
                fvs = (self.history[m]["FittedVals"][s])
                sns.lineplot(
                    x=self.current_dates.to_list()[-len(fvs) :],
                    y=fvs,
                    label=f"{s} {m}",
                    linestyle="--",
                    alpha=0.7,
                    color=color,
                    ax=ax,
                )
                k += 1

    def export(
        self,
        dfs:ExportOptions|list[ExportOptions]=[
            "model_summaries",
            "lvl_test_set_predictions",
            "lvl_fcsts",
        ],
        models:ModelValues="all",
        series:SeriesValues="all",
        cis:bool=False,
        to_excel:bool=False,
        out_path:os.PathLike=Path.cwd(),
        excel_name:str="results.xlsx",
    ) -> pd.DataFrame|dict[str,pd.DataFrame]:
        """ Exports 1-all of 3 pandas dataframes. Can write to excel with each dataframe on a separate sheet.
        Will return either a dictionary with dataframes as values (df str arguments as keys) or a single dataframe if only one df is specified.

        Args:
            dfs (list-like or str): Default 
                ['model_summaries', 'lvl_test_set_predictions', 'lvl_fcsts'].
                A list or name of the specific dataframe(s) you want returned and/or written to excel.
                Must be one of or multiple of the elements in default.
                Exporting test set predictions only works if all exported models were tested using the same test length.
            models (list-like or str): Default 'all'.
               The forecasted models to plot.
               Name of the model, 'all', or list-like of model names.
               'top_' and None not supported.
            series (list-like or str): Default 'all'.
               The series to plot.
               Name of the series ('y1...', 'series1...' or user-selected name), 'all', or list-like of series names.
            to_excel (bool): Default False.
                Whether to save to excel.
            cis (bool): Default False.
                Whether to export confidence intervals for models in 
                "all_fcsts", "test_set_predictions", "lvl_test_set_predictions", "lvl_fcsts"
                dataframes.
            out_path (str): Default './'.
                The path to save the excel file to (ignored when `to_excel=False`).
            excel_name (str): Default 'results.xlsx'.
                The name to call the excel file (ignored when `to_excel=False`).

        Returns:
            (DataFrame or Dict[str,DataFrame]): Either a single pandas dataframe if one element passed to dfs 
            or a dictionary where the keys match what was passed to dfs and the values are dataframes. 

        >>> results = mvf.export(dfs=['model_summaries','lvl_fcsts'],to_excel=True) # returns a dict
        >>> model_summaries = results['model_summaries'] # returns a dataframe
        >>> lvl_fcsts = results['lvl_fcsts'] # returns a dataframe
        >>> ts_preds = mvf.export('test_set_predictions') # returns a dataframe
        """
        _developer_utils.descriptive_assert(
            isinstance(cis, bool),
            "ValueError",
            f"argument passed to cis must be a bool type, not {type(cis)}",
        )

        match dfs:
            case str():
                dfs = [dfs]
            case obj if hasattr(obj, "__len__") and not obj:
                raise ValueError("No values passed to the dfs argument.")
            case _:
                dfs = list(dfs)

        _dfs_ = ["model_summaries","lvl_test_set_predictions","lvl_fcsts"]
        _bad_dfs_ = [i for i in dfs if i not in _dfs_]
        if _bad_dfs_:
            raise ValueError(f"Values passed to the dfs list must be in {_dfs_}, not {_bad_dfs_}")

        series = self._parse_series(series)
        models = self._parse_models(models, hasattr(self, "best_model"))
        output = {}
        if "model_summaries" in dfs:
            cols = [
                "Estimator",
                "Xvars",
                "HyperParams",
                "Lags",
                "Observations",
                "DynamicallyTested",
                "TestSetLength",
                "ValidationMetric",
                "ValidationMetricValue",
                "OptimizedOn",
                "MetricOptimized",
                "best_model",
            ]
            model_summaries = []
            for s in series:
                for m in models:
                    model_summary_sm = pd.DataFrame({"Series": [s], "ModelNickname": [m]})
                    for c in cols:
                        if c in self.history[m]:
                            attr = self.history[m][c]
                            match attr:
                                case dict():
                                    if c in ['HyperParams','Lags']:
                                        model_summary_sm[c] = [attr]
                                    else:
                                        model_summary_sm[c] = [attr.get(s,pd.NA)]
                                case _:
                                    model_summary_sm[c] = [attr]
                        elif c == "OptimizedOn" and hasattr(self, "best_model"):
                            if self.optimize_on in self.optimizer_funcs:
                                model_summary_sm["OptimizedOn"] = [self.optimize_on]
                            else:
                                series, label = self._parse_series(self.optimize_on)
                                model_summary_sm["OptimizedOn"] = [label]
                            if hasattr(self, "optimize_metric"):
                                model_summary_sm["MetricOptimized"] = self.optimize_metric
                            model_summary_sm["best_model"] = m == self.best_model
                        else:
                            model_summary_sm[c] = pd.NA
                    
                    for c in self.determine_best_by:
                        if c in self.history[m]:
                            match self.history[m][c]:
                                case dict():
                                    model_summary_sm[c] = self.history[m][c][s].score

                    model_summaries.append(model_summary_sm)
            model_summaries = pd.concat(model_summaries)
            output["model_summaries"] = model_summaries
        
        if "lvl_fcsts" in dfs:
            df = pd.DataFrame({"DATE": self.future_dates.to_list()})
            for s in series:
                for m in models:
                    df[f"{s}_{m}_lvl_fcst"] = self.history[m]["Forecast"][s][:]
                    if cis:
                        try:
                            df[f"{s}_{m}_lvl_fcst_upper"] = self.history[m]["UpperCI"][s][:]
                            df[f"{s}_{m}_lvl_fcst_lower"] = self.history[m]["LowerCI"][s][:]
                        except KeyError:
                            _developer_utils._warn_about_not_finding_cis(m)
            output["lvl_fcsts"] = df
        
        if "lvl_test_set_predictions" in dfs:
            if self.test_length > 0:
                df = pd.DataFrame({"DATE": self.current_dates.to_list()[-self.test_length :]})
                
                i = 0
                for s in series:
                    df[f"{s}_actuals"] = self.y[s].to_list()[-self.test_length :]
                    for m in models:
                        df[f"{s}_{m}_lvl_ts"] = self.history[m]["TestSetPredictions"][s][:]
                        if cis:
                            try:
                                df[f"{s}_{m}_lvl_ts_upper"] = self.history[m]["TestSetUpperCI"][s][:]
                                df[f"{s}_{m}_lvl_ts_lower"] = self.history[m]["TestSetLowerCI"][s][:]
                            except KeyError:
                                _developer_utils._warn_about_not_finding_cis(m)
                    i += 1
                output["lvl_test_set_predictions"] = df
            else:
                output["lvl_test_set_predictions"] = pd.DataFrame()
        if to_excel:
            out_path = Path(out_path)
            with pd.ExcelWriter(out_path/excel_name, engine="openpyxl") as writer:
                for k, df in output.items():
                    df.to_excel(writer, sheet_name=k, index=False)

        if len(output.keys()) == 1:
            return list(output.values())[0]
        else:
            return output

    def export_fitted_vals(self, series:SeriesValues="all", models:ModelValues="all") -> pd.DataFrame:
        """ Exports a dataframe of fitted values and actuals.

        Args:
            models (list-like or str): Default 'all'.
               Name of the model, 'all', or list-like of model names.
            series (list-like or str): Default 'all'.
               The series to plot.
               Name of the series ('y1...', 'series1...' or user-selected name), 'all', or list-like of series names.

        Returns:
            (DataFrame): The fitted values for all selected series and models.
        """
        series = self._parse_series(series)
        models = self._parse_models(models, False)
        dfd = {"DATE": self.current_dates.to_list()}
        length = len(dfd["DATE"])
        i = 0
        for s in series:
            dfd[f"{s}_actuals"] = (
                self.y[s].to_list()
            )
            for m in models:
                dfd[f"{s}_{m}_fvs"] = (
                    self.history[m]["FittedVals"][s][:]
                )
                length = min(length, len(dfd[f"{s}_{m}_fvs"]), len(dfd[f"{s}_actuals"]))
            i += 1
        return pd.DataFrame({c: v[-length:] for c, v in dfd.items()})

    def corr(self, train_only:bool=False, disp:Literal['matrix','heatmap']="matrix", df:Optional[pd.DataFrame]=None, **kwargs:Any) -> pd.DataFrame|Figure:
        """ Displays pearson correlation between all stored series in object.

        Args:
            train_only (bool): Default False.
                Whether to only include the training set (to avoid leakage).
            disp (str): One of {'matrix','heatmap'}. Default 'matrix'.
                How to display results.
            df (DataFrame): Optional. A dataframe to display correlation for.
                If specified, a dataframe will be created using all series with no lags.
                This argument exists to make the corr_lags() method work and it is not recommended to use it manually.
            **kwargs: Passed to seaborn.heatmap() function and are ignored if disp == 'matrix'.

        Returns:
            (DataFrame or Figure): The created dataframe if disp == 'matrix' else the heatmap fig.
        """
        if df is None:
            df = pd.DataFrame({k:v.values for k, v in self.y.items()})

        if train_only:
            df = df.iloc[: -self.test_length]

        corr = df.corr()
        if disp == "matrix":
            if len(kwargs) > 0:
                warnings.warn(
                    f"Keyword arguments: {kwargs} ignored when disp == 'matrix'.",
                    category=Warning,
                )
            return corr

        elif disp == "heatmap":
            _, ax = plt.subplots()
            sns.heatmap(corr, **kwargs, ax=ax)
            return ax

        else:
            raise ValueError(f'disp must be one of "matrix","heatmap", got {disp}')

    def corr_lags(self, y:Optional[SeriesName]=None, x:Optional[SeriesName]=None, lags:PositiveInt=1, **kwargs:Any) -> pd.DataFrame:
        """ Displays pearson correlation between one series and another series' lags.

        Args:
            y (str): The series to display as the dependent series. Default will take the first loaded series in the object.
            x (str): The series to display as the independent series. Default will take the second loaded series in the object.
            lags (int): Default 1. The number of lags to display in the independent series.
            **kwargs: Passed to the MVForecaster.corr() method. Will not pass the df arg.

        Returns:
            (DataFrame or Figure): The created dataframe if disp == 'matrix' else the heatmap fig.
        """
        y = self.names[0] if y is None else y
        x = self.names[0] if x is None else x
        series1 = self._parse_series(y)
        series2 = self._parse_series(x)

        df = pd.DataFrame({k:v.to_list() for k, v in self.y.items() if k in (series1[0],series2[0])})

        for i in range(lags):
            df[series2[0] + f"_lag{i+1}"] = df[series2[0]].shift(i + 1)

        df = df.dropna()
        return self.corr(df=df, **kwargs)



from scalecast import GridGenerator

def main():
    GridGenerator.get_example_grids(out_name='ExampleGrids.py',overwrite=True)
    GridGenerator.get_grids('vecm',out_name='VECMGrid.py',overwrite=True)
    GridGenerator.get_mv_grids(overwrite=True)
    GridGenerator.get_empty_grids(overwrite=True)

if __name__ == '__main__':
    main()


from ._utils import _developer_utils  
from ._Forecaster_parent import ForecastError
from .Forecaster import Forecaster
from .types import PositiveInt, NonNegativeInt
from .classes import EvaluatedMetric
from .util import find_seasonal_length
from typing import Literal, Optional, Any
import pandas as pd
import numpy as np
import warnings
import copy
from sklearn.preprocessing import RobustScaler

class SeriesTransformer:
    """ A class for transforming the y attribute in a Forecaster object and reverting it back to its original form.

    Args:
        f (Forecaster): The Forecaster object that will receive each transformation/revert.
        deepcopy (bool): Default True. Whether to store a deepcopy of the Forecaster object in the SeriesTransformer object.
    """
    def __init__(self, f, deepcopy=True):
        self.f = copy.deepcopy(f) if deepcopy else f

    def __repr__(self):
        return "SeriesTransformer(\n{}\n)".format(self.f)

    def __str__(self):
        return self.__repr__()

    def __copy__(self):
        obj = type(self).__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        return obj

    def __deepcopy__(self, memo={}):
        obj = type(self).__new__(self.__class__)
        memo[id(self)] = obj
        for k, v in self.__dict__.items():
            setattr(obj, k, copy.deepcopy(v, memo))
        return obj

    def Transform(self, transform_func:callable, **kwargs:Any) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object.
        
        Args:
            transform_func (function): The function that will be used to make the transformation.
                If using a user function, first argument must be the array to transform.
            **kwargs: Passed to the function passed to transform_func.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> import math
        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> def log10(x):
        >>>     return [math.log(i,base=10) for i in x]
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.Transform(log10)
        """
        self.f.y = pd.Series(transform_func(self.f.y, **kwargs))
        return self.f

    def Revert(self, revert_func:callable, exclude_models:list[str] = [], **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.

        Args:
            revert_func (function): The function that will be used to revert the values.
                If using a user function, first argument must be the array to transform.
            exclude_models (list-like): Models to not revert. This is useful if you are transforming
                and reverting an object over and over.
            **kwargs: Passed to the function passed to revert_func.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> import math
        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> def log10(x):
        >>>     return [math.log(i,base=10) for i in x]
        >>> def log10_revert(x):
        >>>     return [10**i for i in x]
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.Transform(log10)
        >>> f = transformer.Revert(log10_revert)
        """
        self.f.y = pd.Series(revert_func(self.f.y, **kwargs))

        for m, h in self.f.history.items():
            if m in exclude_models:
                continue
            for k in ( 
                "Forecast",
                "TestSetPredictions",
                "TestSetUpperCI",
                "TestSetLowerCI",
                "UpperCI",
                "LowerCI",
                "TestSetActuals",
                "FittedVals",
            ):
                if k in h:
                    h[k] = pd.Series(revert_func(h[k], **kwargs),dtype=float).ffill().to_list()
                elif not k.endswith('CI'):
                    h[k] = []
            for met in self.f.metrics:
                h['TestSet' + met.name.upper()] = EvaluatedMetric(
                    store=met,
                    score=_developer_utils._return_na_if_len_zero(self.f.y.iloc[-len(h["TestSetPredictions"]) :], h["TestSetPredictions"], met.eval_func)
                )
                h['InSample' + met.name.upper()] = EvaluatedMetric(
                    store = met,
                    score = _developer_utils._return_na_if_len_zero(self.f.y.iloc[-len(h['FittedVals']) :], h['FittedVals'], met.eval_func)
                )

        return self.f

    def DetrendTransform(
        self,
        loess:bool = False,
        frac:float = 0.5,
        it:int = 3,
        poly_order:PositiveInt=1,
        ln_trend:bool=False,
        seasonal_lags:NonNegativeInt=0,
        m:int|Literal['auto']='auto',
        fit_intercept:bool=True,
        train_only:bool=False,
    ) -> Forecaster:
        """ Detrends the series using an OLS estimator or using LOESS.
        Only call this once if you want to revert the series later.
        The passed Forecaster object must have future dates or be initiated with `require_future_dates=False`.
        Make sure the test length has already been set as well.
        If the test length changes between the time the transformation is called
        and when models are called, you will get this error when reverting: 
        ValueError: All arrays must be of the same length.
        The ols or lowess model from statsmodels will be stored in the detrend_params attribute with the 'model' key.

        Args:
            loess (bool): Default False. Whether to fit a LOESS curve.
            frac (float): Optional. Default 0.5.
                The fraction of the data used when estimating each y-value.
                A smaller frac value will produce a more rigid trend line, 
                while a larger frac value will produce a smoother trend line.
                Ignored when loess is False. 
            it (int): Optional. Default 3.
                The number of iterations used in the loess algorithm.
                A larger it value will produce a smoother trend line, 
                but may take longer to compute.
                Ignored when loess is False. 
            poly_order (int): Default 1. The polynomial order to use in the fitted trend line.
                Ignored when loess is True. 
            ln_trend (bool): Default False. Whether to use a natural logarithmic trend.
                Ignored when loess is True.
            seasonal_lags (int): Default 0. The number of seasonal lags to use in the estimation.
                Ignored when loess is True.
            m (int or str): Default 'auto'. The number of observations that counts one seasonal step.
                Ignored when seasonal_lags = 0.
                When 'auto', uses the M4 competition values:
                for Hourly: 24, Monthly: 12, Quarterly: 4. Everything else gets inferred if possible.
            fit_intercept (bool): Default True. Whether to fit an intercept in the model.
            train_only (bool): Default False. Whether to fit the LOESS or OLS model on the training set only.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DetrendTransform(ln_trend=True)
        """
        self.detrend_params = {
            'origy':self.f.y.copy(),
            'origdates':self.f.current_dates.copy()
        }

        train_only = train_only if self.f.test_length > 0 else False

        fmod = copy.deepcopy(self.f)
        fmod.drop_all_Xvars()
        fmod.add_time_trend()

        if not loess:
            import statsmodels.api as sm 
            if seasonal_lags > 0:
                if m == 'auto':
                    m = find_seasonal_length(m,fmod.freq)
                    if m == 1:
                        warnings.warn(
                            f'Cannot add seasonal lags automatically for the {fmod.freq} frequency. '
                            'Set a value for m manually.'
                        )
                if m > 1:
                    fmod.add_lagged_terms('t',lags=m*seasonal_lags)
                    fmod.drop_Xvars(*[
                        x for x in fmod.get_regressor_names() if isinstance(x,str) and (
                            x.startswith(
                                'tlag'
                            ) and int(
                                x.split('tlag_')[-1]
                            ) % m != 0
                        ) and x != 't'
                    ])
            if ln_trend:
                fmod.add_logged_terms(*fmod.get_regressor_names(),drop=True)
            fmod.add_poly_terms(*fmod.get_regressor_names(),pwr=poly_order)

            dataset = fmod.export_Xvars_df().set_index('DATE')
            if fit_intercept:
                dataset = sm.add_constant(dataset)

            train_set = dataset.loc[fmod.current_dates].dropna() # full dataset minus future dates
            model_inputs = train_set.iloc[:-fmod.test_length,:] if train_only else train_set.copy() # what will be used to fit the model
            test_set = train_set.iloc[-fmod.test_length:,:] # the test set for reverting models later
            future_set = dataset.loc[fmod.future_dates] # the future inputs for reverting models later

            y = fmod.y.values
            y_train = y.copy() if not train_only else y[:-fmod.test_length].copy()

            ols_mod = sm.OLS(y_train[-len(model_inputs):],model_inputs).fit()
            fvs = ols_mod.predict(train_set) # reverts fitted values
            fvs_fut = ols_mod.predict(future_set) # reverts forecasts
            fvs_test = ols_mod.predict(test_set) # reverts test-set predictions
            self.f.keep_smaller_history(len(train_set))
            self.detrend_params['model'] = ols_mod
            self.detrend_params['fvs'] = fvs.values
            self.detrend_params['fvs_fut'] = fvs_fut
            self.detrend_params['fvs_test'] = fvs_test
        else:
            from statsmodels.nonparametric.smoothers_lowess import lowess
            y = fmod.y.values if not train_only else fmod.y.values[-fmod.test_length:]
            loess_fit = lowess(
                y, 
                fmod.current_xreg['t'][:len(y)], 
                frac=frac, 
                it=it,
            )
            self.detrend_params['model'] = loess_fit
            self.detrend_params['fvs'] = np.interp(
                fmod.current_xreg['t'], 
                loess_fit[:, 0], 
                loess_fit[:, 1], 
            )
            self.detrend_params['fvs_test'] = (
                self.detrend_params['fvs'][-self.f.test_length:] 
                if self.f.test_length > 0 
                else []
            )
            self.detrend_params['fvs_fut'] = np.interp(
                fmod.future_xreg['t'], 
                loess_fit[:, 0], 
                loess_fit[:, 1], 
            )

        self.f.y = pd.Series(self.f.y.values - self.detrend_params['fvs'],dtype=float)
        return self.f

    def DetrendRevert(self, exclude_models:list[str] = []) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes a detrend transformation has already been called and uses all model information
        already recorded from that transformation to revert.
        If the test length changes in the Forecaster object between the time the transformation is called
        and when models are called, you will get this error when reverting: 
        ValueError: All arrays must be of the same length.

        Args:
            exclude_models (list-like): Models to not revert. This is useful if you are transforming
                and reverting an object multiple times.
        
        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DetrendTransform(ln_trend=True)
        >>> f = transformer.DetrendRevert()
        """
        if not hasattr(self,'detrend_params'):
            raise ForecastError('Before reverting a trend, make sure DetrendTransform has already been called.')

        self.f.y = self.detrend_params['origy']
        self.f.current_dates = self.detrend_params['origdates']

        fvs = {
            "TestSetPredictions":self.detrend_params['fvs_test'],
            "TestSetActuals":self.detrend_params['fvs_test'],
            "FittedVals":self.detrend_params['fvs'],
            "Forecast":self.detrend_params['fvs_fut'],
            "TestSetUpperCI":self.detrend_params['fvs_test'],
            "TestSetLowerCI":self.detrend_params['fvs_test'],
            "UpperCI":self.detrend_params['fvs_fut'],
            "LowerCI":self.detrend_params['fvs_fut'],
        }

        for m, h in self.f.history.items():
            if m in exclude_models:
                continue
            for k,v in fvs.items():
                if k in h:
                    h[k] = [i + fvs for i, fvs in zip(h[k],v)]
                elif not k.endswith('CI'):
                    h[k] = []

        #delattr(self,'detrend_params')
        return self.Revert(lambda x: x)  # call here to assign correct test-set metrics
        
    def LogTransform(self) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object using a natural log transformation.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        """
        return self.Transform(np.log)

    def LogRevert(self, **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes a natural log transformation has already been called.

        Args:
            **kwargs: Passed to Transformer.Revert() - 
                arg `exclude_models` accepted here.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.LogTransform()
        >>> f = transformer.LogRevert()
        """
        return self.Revert(np.exp, **kwargs)

    def SqrtTransform(self) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object using a square-root transformation.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.SqrtTransform()
        """
        return self.Transform(np.sqrt)

    def SqrtRevert(self, **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes a square-root transformation has already been called.

        Args:
            **kwargs: Passed to Transformer.Revert() - 
                arg `exclude_models` accepted here.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.SqrtTransform()
        >>> f = transformer.SqrtRevert()
        """
        return self.Revert(np.square, **kwargs)

    def ScaleTransform(self,train_only:bool=False) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object using a scale transformation.
        Scale defined as (array[i] - array.mean()) / array.std().

        Args:
            train_only (bool): Default False.
                Whether to fit the scale transformer on the training set only.
        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.ScaleTransform()
        """
        if hasattr(self, "orig_mean"):
            return

        train_only = train_only if self.f.test_length > 0 else False
        stop_at = len(self.f.y) if not train_only else len(self.f.y) - self.f.test_length

        self.orig_mean = self.f.y.values[:stop_at].mean()
        self.orig_std = self.f.y.values[:stop_at].std()

        def func(x, mean, std):
            return [(i - mean) / std for i in x]

        return self.Transform(func, mean=self.orig_mean, std=self.orig_std)

    def ScaleRevert(self, **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes the scale transformation has been called on the object at some point.
        Revert function: array.std()*array[i]+array.mean().

        Args:
            **kwargs: Passed to Transformer.Revert() - 
                arg `exclude_models` accepted here.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> f.set_test_length(.2) # specify a test set to not leak data with this func
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.ScaleTransform(train_only=True)
        >>> f = transformer.ScaleRevert()
        """

        def func(x, mean, std):
            return [std * i + mean for i in x]

        try:
            f = self.Revert(func, mean=self.orig_mean, std=self.orig_std, **kwargs)
            #delattr(self, "orig_mean")
            #delattr(self, "orig_std")
            return f
        except AttributeError:
            raise ValueError("Cannot revert a series that was never scaled.")

    def RobustScaleTransform(self,train_only:bool=False,**kwargs:Any) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object using a robust scale transformation.
        See the function from scikit-learn: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html#sklearn.preprocessing.RobustScaler.

        Args:
            train_only (bool): Default False.
                Whether to fit the transformer on the training set only.
            **kwargs: Passed to the scikit-learn function.
        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.RobustScaleTransform()
        """

        train_only = train_only if self.f.test_length > 0 else False
        stop_at = len(self.f.y) if not train_only else len(self.f.y) - self.f.test_length

        y = self.f.y.values[:stop_at].reshape(-1,1)

        transformer = RobustScaler(**kwargs).fit(y)
        self.rs_transformer = transformer

        return self.Transform(_developer_utils._reshape_func_input,func=transformer.transform)

    def RobustScaleRevert(self, **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes the scale transformation has been called on the object at some point.

        Args:
            **kwargs: Passed to Transformer.Revert() - 
                arg `exclude_models` accepted here.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> f.set_test_length(.2) # specify a test set to not leak data with this func
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.ScaleTransform(train_only=True)
        >>> f = transformer.ScaleRevert()
        """

        if not hasattr(self,'rs_transformer'):
            raise ValueError("Cannot revert a series that was never scaled with RobustScaleTransform.")

        func = self.rs_transformer.inverse_transform

        return self.Revert(_developer_utils._reshape_func_input,func=func,**kwargs)

    def MinMaxTransform(self,train_only:bool=False) -> Forecaster:
        """ Transforms the y attribute in the Forecaster object using a min-max scale transformation.
        Min-max scale defined as (array[i] - array.min()) / (array.max() - array.min()).

        Args:
            train_only (bool): Default False.
                Whether to fit the minmax transformer on the training set only.
        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.MinMaxTransform()
        """
        if hasattr(self, "orig_min"):
            return

        train_only = train_only if self.f.test_length > 0 else False
        stop_at = len(self.f.y) if not train_only else len(self.f.y) - self.f.test_length

        self.orig_min = self.f.y.values[:stop_at].min()
        self.orig_max = self.f.y.values[:stop_at].max()

        def func(x, amin, amax):
            return [(i - amin) / (amax - amin) for i in x]

        return self.Transform(func, amin=self.orig_min, amax=self.orig_max,)

    def MinMaxRevert(self, **kwargs:Any) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Assumes the min-max scale transformation has been called on the object at some point.
        Revert function: array[i]*(array.max() - array.min()) + array.min().

        Args:
            **kwargs: {assed to Transformer.Revert() - 
                arg `exclude_models` accepted here.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> f.set_test_length(.2) # specify a test set to not leak data with this func
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.MinMaxTransform(train_only=True)
        >>> f = transformer.MinMaxRevert()
        """

        def func(x, amin, amax):
            return [i * (amax - amin) + amin for i in x]

        try:
            f = self.Revert(func, amin=self.orig_min, amax=self.orig_max, **kwargs)
            #delattr(self, "orig_min")
            #delattr(self, "orig_max")
            return f
        except AttributeError:
            raise ValueError("Cannot revert a series that was never scaled.")

    def DiffTransform(self, m:int=1) -> Forecaster:
        """ Takes differences or seasonal differences in the Forecaster object's y attribute.
        If using this transformation, call `Forecaster.add_diffed_terms()` and 
        `Forecaster.add_lagged_terms()` if you want to use those before calling this function.
        Call `Forecaster.add_ar_terms()` and `Forecaster.add_AR_terms()` after calling this function.
        Call twice with the same value of m to take second differences.

        Args:
            m (int): Default 1. The seasonal difference to take.
                1 will difference once. 
                12 will take a seasonal difference assuming 12 periods makes a season (monthly data).
                Any int available.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DiffTransform(1)  # first difference
        >>> f = transformer.DiffTransform(1)  # second difference
        >>> f = transformer.DiffTransform(12) # first 12-period difference
        """
        n = (
            0
            if not hasattr(self, f"orig_y_{m}_0")
            else max(
                [
                    int(a.split("_")[-1]) + 1
                    for a in dir(self)
                    if a.startswith(f"orig_y_{m}")
                ]
            )
        )

        setattr(self, f"orig_y_{m}_{n}", self.f.y.to_list())
        setattr(self, f"orig_dates_{m}_{n}", self.f.current_dates.to_list())

        func = lambda x, m: pd.Series(x).diff(m)
        f = self.Transform(func, m=m)
        f.keep_smaller_history(len(f.y) - m)
        return f

    def DiffRevert(self, m:int=1, exclude_models:list[str] = [], delete_attributes:bool=True) -> Forecaster:
        """ Reverts the y attribute in the Forecaster object, along with all model results.
        Calling this makes so that AR values become unusable and have to be re-added to the object.

        Args:
            m (int): Default 1. The seasonal difference to revert.
                1 will undifference once. 
                12 will undifference seasonally 12 periods (monthly data).
                Any int available. Use the same values to revert as you used
                to transform the object originally.
            exclude_models (list-like): Models to not revert. This is useful if you are transforming
                and reverting an object over and over.
            delete_attributes (bool): Default True. Whether to delete the attributes created in DiffTransform after reverting.
                Deleting the attributes allows you to second difference any series.
                Keeping the attributes makes the object more reusable.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DiffTransform(1)  # first difference
        >>> f = transformer.DiffTransform(1)  # second difference
        >>> f = transformer.DiffTransform(12) # first 12-period difference
        >>> # call revert funcs in reverse order
        >>> f = transformer.DiffRevert(12)
        >>> f = transformer.DiffRevert(1)
        >>> f = transformer.DiffRevert(1) # back to normal
        """

        def seasrevert(transformed, orig, m):
            orig = list(orig)[:m]
            transformed = list(transformed)
            for i in range(len(orig)):
                transformed.insert(i, orig[i])
            for i in range(m, len(transformed)):
                transformed[i] = transformed[i - m] + transformed[i]
            return transformed

        n = max(
            [int(a.split("_")[-1]) for a in dir(self) if a.startswith(f"orig_y_{m}")]
        )

        y_orig = getattr(self, f"orig_y_{m}_{n}")
        dates_orig = getattr(self, f"orig_dates_{m}_{n}")

        self.f.y = pd.Series(seasrevert(self.f.y, y_orig, m))
        self.f.current_dates = pd.Series(dates_orig)

        for mod, h in self.f.history.items():
            if mod in exclude_models:
                continue
            if hasattr(h['Forecast'],'__len__'):
                h["Forecast"] = list(seasrevert(h["Forecast"], self.f.y[-m:], m))[m:]
                h['FittedVals'] = list(
                    seasrevert(h['FittedVals'],self.f.y.values[-len(h['FittedVals'])-m:], m)
                )[m:]
            if hasattr(h['TestSetPredictions'],'__len__'):
                h["TestSetPredictions"] = list(
                    seasrevert(
                        h["TestSetPredictions"],
                        self.f.y.to_list()[-(m + self.f.test_length) : -self.f.test_length],
                        m,
                    )
                )[m:]
                h["TestSetActuals"] = self.f.y.to_list()[-self.f.test_length :]
            if np.isnan(self.f.history[mod]['CILevel']) or 'TestSetUpperCI' not in h: # no cis evaluated
                continue
            # undifference cis
            fcst = h["Forecast"]
            test_preds = h["TestSetPredictions"]
            test_actuals = h["TestSetActuals"]
            test_resids = np.abs([p - a for p, a in zip(test_preds,test_actuals)])
            ci_range = np.percentile(test_resids, 100 * self.f.cilevel)
            self.f._set_cis(
                "TestSetUpperCI",
                "TestSetLowerCI",
                m = mod,
                ci_range = ci_range,
                preds = test_preds,
            )
            self.f._set_cis(
                "UpperCI",
                "LowerCI",
                m = mod,
                ci_range = ci_range,
                preds = fcst,
            )

        if delete_attributes:
            delattr(self, f"orig_y_{m}_{n}")
            delattr(self, f"orig_dates_{m}_{n}")
        return self.Revert(lambda x: x, exclude_models = exclude_models)  # call here to assign correct test-set metrics

    def DeseasonTransform(
        self,
        m:Optional[int] = None,
        model:Literal["additive", "add", "multiplicative", "mul"]='add',
        extrapolate_trend:Literal['freq']|int='freq',
        train_only:bool = False,
        **kwargs:Any,
    ) -> Forecaster:
        """ Deseasons a series using the moving average method offered by statsmodel through the seasonal_decompose() function.

        Args:
            m (int): The number of observations that counts one seasonal step.
                If not specified, will use the inferred seasonality from statsmodels.
            model (str): Default 'add'. One of {"additive", "add", "multiplicative", "mul"}.
                The type of seasonal component.
            extrapolate_trend (str or int): Default 'freq'. If set to > 0, the trend resulting from the convolution is
                linear least-squares extrapolated on both ends (or the single one
                if two_sided is False) considering this many (+1) closest points.
                If set to 'freq', use `freq` closest points. Setting this parameter
                results in no NaN values in trend or resid components.
            train_only (bool): Default False. Whether to fit the seasonal decomposition model on the training set only.
            **kwargs: Passed to seasonal_decompose() function from statsmodels.
                See https://www.statsmodels.org/dev/generated/statsmodels.tsa.seasonal.seasonal_decompose.html.

        Returns:
            (Forecaster): A Forecaster object with the transformed attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DeseasonTransform(model='mul')  # multiplicative deseasoning
        """
        _developer_utils.descriptive_assert(
            model in ("additive", "add", "multiplicative", "mul"),
            ValueError,
            f'The value passed to the model argument must be one of ("additive", "add", "multiplicative", "mul"), got {model}.'
        )
        train_only = train_only if self.f.test_length > 0 else False
        if m is None:
            m = find_seasonal_length('auto',self.f.freq)
        decomp_res = self.f.seasonal_decompose(
            model=model,
            extrapolate_trend=extrapolate_trend,
            period = m,
            train_only=train_only,
            **kwargs,
        )
        current_seasonality = decomp_res.seasonal # check
        f = Forecaster(
            y = current_seasonality,
            current_dates = current_seasonality.index,
            future_dates = len(self.f.future_dates) + (self.f.test_length if train_only else 0)
        )
        f.set_estimator('naive')
        f.manual_forecast(
            seasonal=True,
            m=m,
            test_model=False,
            bank_history=False,
        )
        params = {
            'model':model,
            'y':self.f.y.copy(),
            'current_seasonality':current_seasonality.to_list() + f.forecast[:-len(self.f.future_dates)],
            'future_seasonality':f.forecast[-len(self.f.future_dates):],
        }
        setattr(self,f'deseason_params_{m}',params)
        self.f.y = pd.Series(
            self.f.y.values - np.array(params['current_seasonality']) if model in ('add','additive') 
            else self.f.y.values / np.array(params['current_seasonality']),
            dtype=float,
        )
        return self.f

    def DeseasonRevert(self, m:Optional[int] = None, exclude_models:list[str] = []) -> Forecaster:
        """ Reverts a seasonal adjustment already taken on the series. Call DeseasonTransform() before calling this.

        Args:
            m (int): The number of observations that counts one seasonal step.
                If not specified, will use the inferred seasonality from statsmodels.
            exclude_models (list-like): Models to not revert. This is useful if you are transforming
                and reverting an object multiple times.

        Returns:
            (Forecaster): A Forecaster object with the reverted attributes.

        >>> from scalecast.Forecaster import Forecaster
        >>> from scalecast.SeriesTransformer import SeriesTransformer
        >>> f = Forecaster(...)
        >>> transformer = SeriesTransformer(f)
        >>> f = transformer.DeseasonTransform(model='mul')  # multiplicative deseasoning
        >>> f = transformer.DeseasonRevert() # back to normal
        """
        if m is None:
            from statsmodels.tsa.tsatools import freq_to_period
            m = freq_to_period(self.f.freq)
        params = getattr(self,f'deseason_params_{m}')
        try:
            self.f.y = params['y']
        except AttributeError:
            raise ForecastError(
                f'Before reverting a seasonal transformation, make sure DeseasonTransform with m = {m} has already been called.'
            )

        atts = {
            "TestSetPredictions":params['current_seasonality'],
            "TestSetActuals":params['current_seasonality'],
            "FittedVals":params['current_seasonality'],
            "Forecast":params['future_seasonality'],
            "TestSetUpperCI":params['current_seasonality'],
            "TestSetLowerCI":params['current_seasonality'],
            "UpperCI":params['future_seasonality'],
            "LowerCI":params['future_seasonality'],
        }

        for mod, h in self.f.history.items():
            if mod in exclude_models:
                continue
            for a, v in atts.items():
                if a in h:
                    h[a] = [
                        (i + s) if params['model'] in ('add','additive') 
                        else (i * s) for i, s in zip(h[a],v[-len(h[a]):])
                    ]
                elif not a.endswith('CI'):
                    h[a] = []

        #delattr(self, f"deseason_params_{m}")
        return self.Revert(lambda x: x)  # call here to assign correct test-set metrics